@@ -22,7 +22,7 @@
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
-   $Date: 2014-05-27 16:25:29 +0100 (Tue, 27 May 2014) $ $Revision: 37914 $
+   $Date: 2014-10-06 12:35:46 +0100 (Mon, 06 Oct 2014) $ $Revision: 38108 $
    */ 
 
 #import	<Foundation/NSString.h>
@@ -170,10 +170,17 @@ connectQuote(NSString *str)
 	      [m appendString: @" password="];
 	      [m appendString: str];
 	    }
+	  str = connectQuote([self clientName]);
+	  if (str != nil)
+	    {
+	      [m appendString: @" application_name="];
+	      [m appendString: str];
+	    }
 
 	  if ([self debugging] > 0)
 	    {
-	      [self debug: @"Connect to '%@' as %@", m, [self name]];
+	      [self debug: @"Connect to '%@' as %@ (%@)",
+                m, [self name], [self clientName]];
 	    }
 	  connection = PQconnectdb([m UTF8String]);
 	  if (PQstatus(connection) != CONNECTION_OK)
@@ -194,18 +201,38 @@ connectQuote(NSString *str)
 	    {
 	      const char	*p;
 
+              backendPID = PQbackendPID(connection);
+
+	      connected = YES;
+
 	      p = PQparameterStatus(connection, "standard_conforming_strings");
               if (p != 0)
                 {
+                  /* The standard conforming strings setting exists,
+                   * so the E'...' syntax must exist and we can use
+                   * it for byte arrays.
+                   */
                   escapeStrings = YES;
+
+                  /* If the escape_string_warning setting is on,
+                   * the server will warn about backslashes even
+                   * in properly quoted strings, so turn it off.
+                   */
+                  if (strcmp(p, "on") == 0)
+                    {
+                      [self execute: @"SET escape_string_warning=off", nil];
+                    }
+                  else
+                    {
+                      [self execute: @"SET standard_conforming_strings=on;"
+                        @"SET escape_string_warning=off", nil];
+                    }
                 }
               else
                 {
                   escapeStrings = NO;
                 }
-              backendPID = PQbackendPID(connection);
 
-	      connected = YES;
 	      if ([self debugging] > 0)
 		{
 		  [self debug: @"Connected to '%@'", [self name]];
@@ -380,7 +407,6 @@ connectQuote(NSString *str)
             {
               str = [NSString stringWithCString: cstr];
             }
-	  [self disconnect];
 	  [NSException raise: SQLException format: @"Error executing %@: %@",
 	    stmt, str];
 	}
@@ -400,14 +426,14 @@ connectQuote(NSString *str)
     {
       NSString	*n = [localException name];
 
-      if ([n isEqual: SQLConnectionException] == YES) 
-	{
-	  [self disconnect];
-	}
       if (result != 0)
 	{
 	  PQclear(result);
 	}
+      if (YES == connected && [n isEqual: SQLConnectionException] == YES) 
+	{
+	  [self disconnect];
+	}
       [localException retain];
       [arp release];
       [localException autorelease];
@@ -513,7 +539,6 @@ static unsigned int trim(char *str)
             {
               str = [NSString stringWithCString: cstr];
             }
-	  [self disconnect];
 	  [NSException raise: SQLException format: @"Error executing %@: %@",
 	    stmt, str];
 	}
@@ -630,14 +655,14 @@ static unsigned int trim(char *str)
     {
       NSString	*n = [localException name];
 
-      if ([n isEqual: SQLConnectionException] == YES) 
-	{
-	  [self disconnect];
-	}
       if (result != 0)
 	{
 	  PQclear(result);
 	}
+      if (YES == connected && [n isEqual: SQLConnectionException] == YES) 
+	{
+	  [self disconnect];
+	}
       [records release];
       records = nil;
       [localException retain];
@@ -981,7 +1006,10 @@ static unsigned int trim(char *str)
 {
   if (extra != 0)
     {
-      [self disconnect];
+      if (YES == connected)
+        {
+          [self disconnect];
+        }
       NSZoneFree(NSDefaultMallocZone(), extra);
     }
   [super dealloc];
