@@ -172,7 +172,7 @@
   </section>
 </chapter>
 
-   $Date: 2014-05-27 16:22:35 +0100 (Tue, 27 May 2014) $ $Revision: 37913 $
+   $Date: 2014-10-13 11:47:06 +0100 (Mon, 13 Oct 2014) $ $Revision: 38114 $
    */ 
 
 #ifndef	INCLUDED_SQLClient_H
@@ -181,6 +181,7 @@
 #import	<Foundation/NSArray.h>
 #import	<Foundation/NSObject.h>
 
+@class	NSConditionLock;
 @class	NSCountedSet;
 @class	NSMapTable;
 @class	NSMutableDictionary;
@@ -341,6 +342,8 @@ extern NSTimeInterval	SQLClientTimeStart();
  */
 extern unsigned	SQLClientTimeTick();
 
+@class SQLClientPool;
+
 /**
  * <p>The SQLClient class encapsulates dynamic SQL access to relational
  * database systems.  A shared instance of the class is used for
@@ -382,20 +385,22 @@ SQLCLIENT_PRIVATE
   NSString		*_user;		/** The configured user */
   NSMutableArray	*_statements;	/** Uncommitted statements */
   /**
-   * Timestamp of last operation.<br />
+   * Timestamp of completion of last operation.<br />
    * Maintained by -simpleExecute: -simpleQuery:recordType:listType:
    * and -cache:simpleQuery:recordType:listType:
    * Also set for a failed connection attempt, but not reported by the
    * -lastOperation method in that case.
    */
   NSTimeInterval	_lastOperation;	
-  NSTimeInterval	_duration;
+  NSTimeInterval	_lastStart;	/** Last op start or connect */
+  NSTimeInterval	_duration;      /** Duration logging threshold */
   unsigned int		_debugging;	/** The current debugging level */
   GSCache		*_cache;	/** The cache for query results */
   NSThread		*_cacheThread;	/** Thread for cache queries */
   unsigned int		_connectFails;	/** The count of connection failures */
   NSMapTable            *_observers;    /** Observations of async events */
   NSCountedSet          *_names;        /** Track notification names */
+  SQLClientPool         *_pool;         /** The pool of the client (or nil) */
   /** Allow for extensions by allocating memory and pointing to it from
    * the _extra ivar.  That way we can avoid binary incompatibility between
    * minor releases.
@@ -424,8 +429,9 @@ SQLCLIENT_PRIVATE
 + (SQLClient*) existingClient: (NSString*)reference;
 
 /**
- * Return the maximum number of simultaneous database connections
- * permitted (set by +setMaxConnections: and defaults to 8)
+ * Returns the maximum number of simultaneous database connections
+ * permitted (set by +setMaxConnections: and defaults to 8) for
+ * connections outside of SQLClientPool instances.
  */
 + (unsigned int) maxConnections;
 
@@ -442,6 +448,9 @@ SQLCLIENT_PRIVATE
  * idle connections, but you can also pass a date in the future to
  * close all connections.
  * </p>
+ * <p>Purging does not apply to connections made by SQLClientPool
+ * instances.
+ * </p>
  */
 + (void) purgeConnections: (NSDate*)since;
 
@@ -452,6 +461,9 @@ SQLCLIENT_PRIVATE
  * <p>This value is used by the +purgeConnections: method to determine how
  * many connections should be disconnected when it is called.
  * </p>
+ * <p>Connections used by SQLClientPool instances are not considered by
+ * this maximum.
+ * </p>
  */
 + (void) setMaxConnections: (unsigned int)c;
 
@@ -563,8 +575,8 @@ SQLCLIENT_PRIVATE
 
 /**
  * Return a flag to say whether a connection to the database server is
- * currently live.  This is mostly useful for debug/reporting, but is
- * used internally to keep track of active connections.
+ * currently live (the value of the 'connected' instance variable).<br />
+ * This is mostly useful for debug/reporting.
  */
 - (BOOL) connected;
 
@@ -622,6 +634,13 @@ SQLCLIENT_PRIVATE
  */
 - (id) initWithConfiguration: (NSDictionary*)config;
 
+/**
+ * Calls -initWithConfiguration:name:pool: passing NO to say the client is
+ * not in a pool.
+ */
+- (id) initWithConfiguration: (NSDictionary*)config
+			name: (NSString*)reference;
+
 /**
  * Initialise using the supplied configuration, or if that is nil, try to
  * use values from NSUserDefaults (and automatically update when the
@@ -630,10 +649,10 @@ SQLCLIENT_PRIVATE
  * a nil name is supplied, defaults to the value of SQLClientName in the
  * configuration dictionary (or in the standard user defaults).  If there is
  * no value for SQLClientName, uses the string 'Database'.<br />
- * If a SQLClient instance already exists with the name used for this
- * instance, the receiver is deallocated and the existing instance is
- * retained and returned ... there may only ever be one instance for a
- * particular reference name.<br />
+ * If pool is nil and a SQLClient instance already exists with the
+ * name used for this instance, the receiver is deallocated and the existing
+ * instance is retained and returned ... there may only ever be one instance
+ * for a particular reference name which is not in a pool.<br />
  * <br />
  * The config argument (or the SQLClientReferences user default)
  * is a dictionary with names as keys and dictionaries
@@ -653,7 +672,8 @@ SQLCLIENT_PRIVATE
  * connect to a database on a different host over the network.
  */
 - (id) initWithConfiguration: (NSDictionary*)config
-			name: (NSString*)reference;
+			name: (NSString*)reference
+                        pool: (SQLClientPool*)pool;
 
 /** Two clients are considered equal if they refer to the same database
  * and are logged in as the same database user using the same protocol.
@@ -678,6 +698,14 @@ SQLCLIENT_PRIVATE
  */
 - (NSDate*) lastOperation;
 
+/** Compares the receiver with the other client to see which one has been
+ * inactive but connected for longest (if they are connected) and returns
+ * that instance.<br />
+ * If neither is idle but connected, the method returns nil.<br />
+ * In a tie, the method returns the other instance.
+ */
+- (SQLClient*) longestIdle: (SQLClient*)other;
+
 /**
  * Return the database reference name for this instance (or nil).
  */
@@ -1367,13 +1395,132 @@ SQLCLIENT_PRIVATE
  * asynchronous query to update them will be run on the cache thread.<br />
  * The rule is that, if the item's age is more than twice its nominal
  * lifetime, it will be retrieved immediately, otherwise it will be
- * retrieved asynchrnonously.<br />
+ * retrieved asynchronously.<br />
  * Currently this may only be the main thread or nil.  Any attempt to set
  * another thread will use the main thread instead.
  */
 - (void) setCacheThread: (NSThread*)aThread;
 @end
 
+/** <p>An SQLClientPool instance may be used to create/control a pool of
+ * client objects.  Code may obtain autoreleased proxies to the clients
+ * from the pool and use them safe in the knowledge that they won't be
+ * used anywhere else ... as soon as the client would be deallocated, it
+ * is returned to the pool.
+ * </p>
+ * <p>All clients in the pool share the same cache object, so query results
+ * cached by one client will be available to other clients in the pool.
+ * </p>
+ */
+@interface	SQLClientPool : NSObject
+{
+  NSConditionLock       *lock;  /** Controls access to the pool contents */
+  SQLClient             **c;    /** The clients of the pool. */
+  BOOL                  *u;     /** Whether the client is in use. */
+  int                   max;    /** Maximum connection count */
+  int                   min;    /** Minimum connection count */
+  NSDictionary          *_config;       /** The pool configuration object */
+  NSString              *_name;         /** The pool configuration name */
+  NSTimeInterval	_duration;      /** Duration logging threshold */
+  unsigned int		_debugging;	/** The current debugging level */
+  uint64_t              _immediate;     /** Immediate client provisions */
+  uint64_t              _delayed;       /** Count of delayed provisions */
+  uint64_t              _failed;        /** Count of timed out provisions */
+  NSTimeInterval        _longest;       /** Count of longest delay */
+  NSTimeInterval        _delayWaits;    /** Time waiting for provisions */
+  NSTimeInterval        _failWaits;     /** Time waiting for timewouts */
+}
+
+/** Returns the count of currently available connections in the pool.
+ */
+- (int) availableConnections;
+
+/**
+ * Creates a pool of clients using a single client configuration.<br />
+ * Calls -initWithConfiguration:name:pool: (passing NO to say the client
+ * is not in a pool) to create each client.<br />
+ * The value of maxConnections is the size of the pool (ie the number of
+ * clients created) and thus the maximum number of concurrent connections
+ * to the database server.<br />
+ * The value of minConnections is the minimum number of connected clients
+ * normally expected to be in the pool.  The pool tries to ensure that it
+ * doesn't contain more than this number of idle connected clients.
+ */
+- (id) initWithConfiguration: (NSDictionary*)config
+			name: (NSString*)reference
+                         max: (int)maxConnections
+                         min: (int)minConnections;
+
+/** Return the maximum number of database connections in the pool.
+ */
+- (int) maxConnections;
+
+/** Return the minimum number of database connections in the pool.
+ */
+- (int) minConnections;
+
+/** Fetches an (autoreleased) client from the pool.<br />
+ * This method blocks indefinitely waiting for a client to become
+ * available in the pool.
+ */
+- (SQLClient*) provideClient;
+
+/** Fetches an (autoreleased) client from the pool.<br />
+ * If no client is or becomes available before the specified date then
+ * the method returns nil.<br />
+ * If when is nil then a date in the distant future is used so that
+ * the method will effectively wait forever to get a client.
+ */
+- (SQLClient*) provideClientBeforeDate: (NSDate*)when;
+
+/**
+ * Sets the cache for all the clients in the pool.
+ */
+- (void) setCache: (GSCache*)aCache;
+
+/**
+ * Sets the cache thread for all the clients in the pool.
+ */
+- (void) setCacheThread: (NSThread*)aThread;
+
+/** Set the debugging level for all clients in the pool.
+ */
+- (void) setDebugging: (unsigned int)level;
+
+/** Set the duration logging threshold for all clients in the pool.
+ */
+- (void) setDurationLogging: (NSTimeInterval)threshold;
+
+/** Sets the pool size limits (number of connections we try to maintain).<br />
+ * The value of maxConnections is the size of the pool (ie the number of
+ * clients created) and thus the maximum number of concurrent connections
+ * to the database server.<br />
+ * The value of minConnections is the minimum number of connected clients
+ * normally expected to be in the pool.  The pool tries to ensure that it
+ * doesn't contain more than this number of idle connected clients.<br />
+ * The value of maxConnections must be greater than or equal to that of
+ * minConnections and may not be greater than 100.
+ * The value of minConnections must be less than or equal to that of
+ * maxConnections and may not be less than 1.
+ */
+- (void) setMax: (int)maxConnections min: (int)minConnections;
+
+/** Returns a string describing the usage of the pool.
+ */
+- (NSString*) statistics;
+
+/** Puts the client back in the pool.  This happens automatically
+ * when a client from a pool would normally be deallocated so you don't
+ * generally need to do it.<br />
+ * Returns YES if the supplied client was from the pool, NO otherwise.<br />
+ * If the swallowed client would take the count of idle client connections
+ * in the pool above the configured minimum, the oldest (ie longest idle)
+ * client in the pool is disconnected.
+ */
+- (BOOL) swallowClient: (SQLClient*)client;
+
+@end
+
 /**
  * The SQLTransaction transaction class provides a convenient mechanism
  * for grouping together a series of SQL statements to be executed as a
@@ -1397,6 +1544,7 @@ SQLCLIENT_PRIVATE
   unsigned		_count;
   BOOL                  _batch;
   BOOL                  _stop;
+  uint8_t               _merge;
 }
 
 /**
@@ -1432,7 +1580,7 @@ SQLCLIENT_PRIVATE
 - (id) copyWithZone: (NSZone*)z;
 
 /**
- * Returns the number of individual statements ond/r subsidiary transactions
+ * Returns the number of individual statements and/or subsidiary transactions
  * which have been added to the receiver.  For a count of the total number
  * of statements, use the -totalCount method.
  */
@@ -1520,6 +1668,38 @@ SQLCLIENT_PRIVATE
  */
 - (void) reset;
 
+/** <p>Use this method to enable merging of statemements subsequently added
+ * or appended to the receiver.  The history argument specifies how many
+ * of the most recent statements in the transaction should be checked for
+ * merging in a new statement, with a value of zero meaning that no
+ * merging is done.<br />
+ * Returns the previous setting for the transaction.
+ * </p>
+ * <p>You may use this feature with an insert statement of the form:<br />
+ * INSERT INTO table (fieldnames) VALUES (values);<br />
+ * For databases which support multiline inserts such that they can be
+ * merged into something of the form:
+ * INSERT INTO table (fieldnames) VALUES (values1),(values2),...;
+ * </p>
+ * <p>Or may use this with an update or delete statement of the form:<br />
+ * command table SET settings WHERE condition;<br />
+ * So that statements may be merged into:<br />
+ * command table SET settings WHERE (condition1) OR (condition2) OR ...;
+ * </p>
+ * If no opportunity for merging is found, the new statement is simply
+ * added to the transaction.<br />
+ * Caveats:<br />
+ * 1. databases may not actually support multiline insert.<br />
+ * 2. Merging is done only if the statement up to the string 'VALUES'
+ * (for insert) or 'WHERE' (for update) matches.<br />
+ * 3. Merging into any of the last N statements (where N is greater than 1)
+ * may of course change the order of statements in the transaction,
+ * so care must be taken not to use this feature where that might matter.<br />
+ * 4. This is a simple text match rather than sql syntactic analysis,
+ * so it's possible to confuse the process with complex statements.
+ */
+- (uint8_t) setMerge: (uint8_t)history;
+
 /**
  * Returns the total count of statements in this transaction including
  * those in any subsidiary transactions.  For a count of the statements
