@@ -22,7 +22,7 @@
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 
-   $Date: 2014-05-19 10:31:02 +0100 (Mon, 19 May 2014) $ $Revision: 37893 $
+   $Date: 2014-10-13 11:47:06 +0100 (Mon, 13 Oct 2014) $ $Revision: 38114 $
    */ 
 
 #import	<Foundation/NSArray.h>
@@ -36,9 +36,9 @@
 #import	<Foundation/NSDictionary.h>
 #import	<Foundation/NSEnumerator.h>
 #import	<Foundation/NSException.h>
-#import	<Foundation/NSHashTable.h>
 #import	<Foundation/NSKeyValueCoding.h>
 #import	<Foundation/NSLock.h>
+#import	<Foundation/NSHashTable.h>
 #import	<Foundation/NSMapTable.h>
 #import	<Foundation/NSNotification.h>
 #import	<Foundation/NSNull.h>
@@ -76,10 +76,11 @@ NSString * const SQLClientDidDisconnectNotification
 static NSNull	*null = nil;
 static NSArray	*queryModes = nil;
 static NSThread	*mainThread = nil;
-static Class	NSStringClass = 0;
-static Class	NSArrayClass = 0;
-static Class	NSDateClass = 0;
-static Class	NSSetClass = 0;
+static Class	NSStringClass = Nil;
+static Class	NSArrayClass = Nil;
+static Class	NSDateClass = Nil;
+static Class	NSSetClass = Nil;
+static Class	SQLClientClass = Nil;
 
 @interface	_ConcreteSQLRecord : SQLRecord
 {
@@ -105,237 +106,6 @@ static Class	NSSetClass = 0;
 }
 @end
 
-@interface	SQLClientPool : NSObject
-{
-  unsigned	pool;
-  NSString	*name;
-  NSString	*serv;
-  NSString	*user;
-  NSString	*pass;
-  NSString	*path;
-  NSHashTable	*idle;
-  NSHashTable	*used;
-}
-- (BOOL) isSingle;
-- (BOOL) makeIdle: (SQLClient*)c;
-- (BOOL) makeUsed: (SQLClient*)c;
-- (void) setConfiguration: (NSDictionary*)o;
-@end
-
-@implementation	SQLClientPool
-- (void) dealloc
-{
-  if (idle != 0)
-    {
-      NSFreeHashTable(idle);
-      idle = 0;
-    }
-  if (used != 0)
-    {
-      NSFreeHashTable(used);
-      used = 0;
-    }
-  [name release]; name = nil;
-  [serv release]; serv = nil;
-  [user release]; user = nil;
-  [pass release]; pass = nil;
-  [path release]; path = nil;
-  [super dealloc];
-}
-
-- (id) initWithConfiguration: (NSDictionary*)config
-			name: (NSString*)reference
-{
-  name = [reference copy];
-  idle = NSCreateHashTable(NSNonRetainedObjectHashCallBacks, 16);
-  used = NSCreateHashTable(NSNonRetainedObjectHashCallBacks, 16);
-  [self setConfiguration: config];
-  return self;
-}
-
-- (BOOL) isSingle
-{
-  if (pool == 1)
-    {
-      return YES;
-    }
-  return NO;
-}
-
-- (BOOL) makeIdle: (SQLClient*)c
-{
-  if (NSHashGet(idle, (void*)c) == (void*)c)
-    {
-      return YES;				// Already idle
-    }
-  if (NSHashGet(used, (void*)c) == (void*)c)
-    {
-      NSHashRemove(used, (void*)c);
-    }
-  if (NSCountHashTable(idle) + NSCountHashTable(used) < pool)
-    {
-      NSHashInsert(idle, (void*)c);
-      return YES;
-    }
-  return NO;
-}
-
-- (BOOL) makeUsed: (SQLClient*)c
-{
-  if (NSHashGet(used, (void*)c) == (void*)c)
-    {
-      return YES;				// Already used
-    }
-  if (NSHashGet(idle, (void*)c) == (void*)c)
-    {
-      NSHashRemove(idle, (void*)c);
-    }
-  if (NSCountHashTable(idle) + NSCountHashTable(used) < pool)
-    {
-      NSHashInsert(used, (void*)c);
-      return YES;
-    }
-  return NO;
-}
-
-- (void) setConfiguration: (NSDictionary*)o
-{
-  NSDictionary	*d;
-  NSString	*s;
-  BOOL		change = NO;
-  int		capacity;
-
-  /*
-   * get dictionary containing config info for this client by name.
-   */
-  d = [o objectForKey: @"SQLClientReferences"];
-  if ([d isKindOfClass: [NSDictionary class]] == NO)
-    {
-      d = nil;
-    }
-  d = [d objectForKey: name];
-  if ([d isKindOfClass: [NSDictionary class]] == NO)
-    {
-      d = nil;
-    }
-
-  s = [d objectForKey: @"ServerType"];
-  if ([s isKindOfClass: NSStringClass] == NO)
-    {
-      s = @"Postgres";
-    }
-  if (s != serv && [s isEqual: serv] == NO)
-    {
-      s = [s copy];
-      [serv release];
-      serv = s;
-      change = YES;
-    }
-
-  s = [d objectForKey: @"Database"];
-  if ([s isKindOfClass: NSStringClass] == NO)
-    {
-      s = [o objectForKey: @"Database"];
-      if ([s isKindOfClass: NSStringClass] == NO)
-	{
-	  s = nil;
-	}
-    }
-  if (s != path && [s isEqual: path] == NO)
-    {
-      s = [s copy];
-      [path release];
-      path = s;
-      change = YES;
-    }
-
-  s = [d objectForKey: @"User"];
-  if ([s isKindOfClass: NSStringClass] == NO)
-    {
-      s = [o objectForKey: @"User"];
-      if ([s isKindOfClass: NSStringClass] == NO)
-	{
-	  s = @"";
-	}
-    }
-  if (s != user && [s isEqual: user] == NO)
-    {
-      s = [s copy];
-      [user release];
-      user = s;
-      change = YES;
-    }
-
-  s = [d objectForKey: @"Password"];
-  if ([s isKindOfClass: NSStringClass] == NO)
-    {
-      s = [o objectForKey: @"Password"];
-      if ([s isKindOfClass: NSStringClass] == NO)
-	{
-	  s = @"";
-	}
-    }
-  if (s != pass && [s isEqual: pass] == NO)
-    {
-      s = [s copy];
-      [pass release];
-      pass = s;
-      change = YES;
-    }
-
-  s = [d objectForKey: @"Password"];
-  if ([s isKindOfClass: NSStringClass] == NO)
-    {
-      s = @"1";
-    }
-  capacity = [s intValue];
-  if (capacity < 1) capacity = 1;
-  if (capacity > 100) capacity = 100;
-
-  if (change == YES)
-    {
-      NSResetHashTable(idle);
-      NSResetHashTable(used);
-    }
-  if (pool > capacity)
-    {
-      unsigned	ic = NSCountHashTable(idle);
-      unsigned	uc = NSCountHashTable(used);
-
-      if (ic + uc > capacity)
-        {
-	  NSHashEnumerator	e = NSEnumerateHashTable(idle);
-	  void			*c;
-
-	  while (ic + uc > capacity
-	    && (c = NSNextHashEnumeratorItem(&e)) != nil)
-	    {
-	      NSHashRemove(idle, c);
-	      ic--;
-	    }
-	  NSEndHashTableEnumeration(&e);
-	  if (uc > capacity)
-	    {
-	      NSHashEnumerator	e = NSEnumerateHashTable(used);
-	      void		*c;
-
-	      while (uc > capacity
-		&& (c = NSNextHashEnumeratorItem(&e)) != nil)
-		{
-		  NSHashRemove(used, c);
-		  uc--;
-		}
-	      NSEndHashTableEnumeration(&e);
-	    }
-        }
-    }
-  pool = capacity;
-}
-
-@end
-
-
-
 static Class aClass = 0;
 static Class rClass = 0;
 
@@ -866,8 +636,9 @@ static NSTimeInterval	classDuration = -1;
 /**
  * Container for all instances.
  */
+static NSHashTable	*clientsHash = 0;
 static NSMapTable	*clientsMap = 0;
-static NSRecursiveLock	*clientsMapLock = nil;
+static NSRecursiveLock	*clientsLock = nil;
 static NSString		*beginString = @"begin";
 static NSArray		*beginStatement = nil;
 static NSString		*commitString = @"commit";
@@ -877,12 +648,53 @@ static NSArray		*rollbackStatement = nil;
 
 
 @interface	SQLClient (Private)
+
+/**
+ * Internal method to handle configuration using the notification object.
+ * This object may be either a configuration front end or a user defaults
+ * object ... so we have to be careful that we work with both.
+ */
 - (void) _configure: (NSNotification*)n;
+
+/** Internal method to make the client instance lock available to
+ * an associated SQLTransaction
+ */
+- (NSRecursiveLock*) _lock;
+
+/** Internal method to populate the cache with the result of a query.
+ */
 - (void) _populateCache: (CacheQuery*)a;
-- (NSArray*) _prepare: (NSString*)stmt args: (va_list)args;
+
+/**
+ * Internal method to build an sql string by quoting any non-string objects
+ * and concatenating the resulting strings in a nil terminated list.<br />
+ * Returns an array containing the statement as the first object and
+ * any NSData objects following.  The NSData objects appear in the
+ * statement strings as the marker sequence - <code>'?'''?'</code>
+ */
+- (NSMutableArray*) _prepare: (NSString*)stmt args: (va_list)args;
+
+/** Internal method called to record the 'main' thread in which automated
+ * cache updates are to be performed.
+ */
 - (void) _recordMainThread;
-- (NSArray*) _substitute: (NSString*)str with: (NSDictionary*)vals;
+
+/**
+ * Internal method to substitute values from the dictionary into
+ * a string containing markup identifying where the values should
+ * appear by name.  Non-string objects in the dictionary are quoted.<br />
+ * Returns an array containing the statement as the first object and
+ * any NSData objects following.  The NSData objects appear in the
+ * statement strings as the marker sequence - <code>'?'''?'</code>
+ */
+- (NSMutableArray*) _substitute: (NSString*)str with: (NSDictionary*)vals;
+
+/*
+ * Called at one second intervals to ensure that our current timestamp
+ * is reasonably accurate.
+ */
 + (void) _tick: (NSTimer*)t;
+
 @end
 
 @interface	SQLClient (GSCacheDelegate)
@@ -895,14 +707,23 @@ static NSArray		*rollbackStatement = nil;
 @implementation	SQLClient
 
 static unsigned int	maxConnections = 8;
+static int	        poolConnections = 0;
 
 + (NSArray*) allClients
 {
-  NSArray	*a;
+  NSMutableArray	*a;
+  NSHashEnumerator	e;
+  id                    o;
 
-  [clientsMapLock lock];
-  a = NSAllMapTableValues(clientsMap);
-  [clientsMapLock unlock];
+  [clientsLock lock];
+  a = [NSMutableArray arrayWithCapacity: NSCountHashTable(clientsHash)];
+  e = NSEnumerateHashTable(clientsHash);
+  while (nil != (o = (id)NSNextHashEnumeratorItem(&e)))
+    {
+      [a addObject: o];
+    }
+  NSEndHashTableEnumeration(&e);
+  [clientsLock unlock];
   return a;
 }
 
@@ -951,37 +772,43 @@ static unsigned int	maxConnections = 8;
 	}
     }
 
-  [clientsMapLock lock];
+  [clientsLock lock];
   existing = (SQLClient*)NSMapGet(clientsMap, reference);
   [[existing retain] autorelease];
-  [clientsMapLock unlock];
+  [clientsLock unlock];
   return existing;
 }
 
 + (void) initialize
 {
-  static id	modes[1];
-  
-  modes[0] = NSDefaultRunLoopMode;
-  queryModes = [[NSArray alloc] initWithObjects: modes count: 1];
-  GSTickerTimeNow();
-  [SQLRecord class];	// Force initialisation
-  if (clientsMap == 0)
+  if (Nil == SQLClientClass && [SQLClient class] == self)
     {
-      clientsMap = NSCreateMapTable(NSObjectMapKeyCallBacks,
-        NSNonRetainedObjectMapValueCallBacks, 0);
-      clientsMapLock = [NSRecursiveLock new];
-      beginStatement = [[NSArray arrayWithObject: beginString] retain];
-      commitStatement = [[NSArray arrayWithObject: commitString] retain];
-      rollbackStatement = [[NSArray arrayWithObject: rollbackString] retain];
-      NSStringClass = [NSString class];
-      NSArrayClass = [NSArray class];
-      NSSetClass = [NSSet class];
-      [NSTimer scheduledTimerWithTimeInterval: 1.0
-				       target: self
-				     selector: @selector(_tick:)
-				     userInfo: 0
-				      repeats: YES];
+      static id	modes[1];
+      
+      SQLClientClass = self;
+      modes[0] = NSDefaultRunLoopMode;
+      queryModes = [[NSArray alloc] initWithObjects: modes count: 1];
+      GSTickerTimeNow();
+      [SQLRecord class];	// Force initialisation
+      if (0 == clientsHash)
+        {
+          clientsHash = NSCreateHashTable(NSNonOwnedPointerHashCallBacks, 0);
+          clientsMap = NSCreateMapTable(NSObjectMapKeyCallBacks,
+            NSNonRetainedObjectMapValueCallBacks, 0);
+          clientsLock = [NSRecursiveLock new];
+          beginStatement = [[NSArray arrayWithObject: beginString] retain];
+          commitStatement = [[NSArray arrayWithObject: commitString] retain];
+          rollbackStatement
+            = [[NSArray arrayWithObject: rollbackString] retain];
+          NSStringClass = [NSString class];
+          NSArrayClass = [NSArray class];
+          NSSetClass = [NSSet class];
+          [NSTimer scheduledTimerWithTimeInterval: 1.0
+                                           target: self
+                                         selector: @selector(_tick:)
+                                         userInfo: 0
+                                          repeats: YES];
+        }
     }
 }
 
@@ -992,21 +819,20 @@ static unsigned int	maxConnections = 8;
 
 + (void) purgeConnections: (NSDate*)since
 {
-  NSMapEnumerator	e;
-  NSString		*n;
+  NSHashEnumerator	e;
   SQLClient		*o;
   unsigned int		connectionCount = 0;
   NSTimeInterval	t = [since timeIntervalSinceReferenceDate];
 
-  [clientsMapLock lock];
-  e = NSEnumerateMapTable(clientsMap);
-  while (NSNextMapEnumeratorPair(&e, (void**)&n, (void**)&o) != 0)
+  [clientsLock lock];
+  e = NSEnumerateHashTable(clientsHash);
+  while (nil != (o = (SQLClient*)NSNextHashEnumeratorItem(&e)))
     {
       if (since != nil)
 	{
 	  NSTimeInterval	when = o->_lastOperation;
 
-	  if (when < t)
+	  if (when < t && YES == o->connected)
 	    {
 	      [o disconnect];
 	    }
@@ -1016,20 +842,20 @@ static unsigned int	maxConnections = 8;
 	  connectionCount++;
 	}
     }
-  NSEndMapTableEnumeration(&e);
-  [clientsMapLock unlock];
+  NSEndHashTableEnumeration(&e);
+  [clientsLock unlock];
 
-  while (connectionCount >= maxConnections)
+  while (connectionCount >= (maxConnections + poolConnections))
     {
       SQLClient		*other = nil;
       NSTimeInterval	oldest = 0.0;
   
       connectionCount = 0;
-      [clientsMapLock lock];
-      e = NSEnumerateMapTable(clientsMap);
-      while (NSNextMapEnumeratorPair(&e, (void**)&n, (void**)&o))
+      [clientsLock lock];
+      e = NSEnumerateHashTable(clientsHash);
+      while (nil != (o = (SQLClient*)NSNextHashEnumeratorItem(&e)))
 	{
-	  if ([o connected] == YES)
+	  if (YES == o->connected)
 	    {
 	      NSTimeInterval	when = o->_lastOperation;
 
@@ -1041,13 +867,13 @@ static unsigned int	maxConnections = 8;
 		}
 	    }
 	}
-      NSEndMapTableEnumeration(&e);
-      [clientsMapLock unlock];
+      NSEndHashTableEnumeration(&e);
+      [clientsLock unlock];
       connectionCount--;
       if ([other debugging] > 0)
 	{
 	  [other debug:
-	    @"Force disconnect of '%@' because pool size (%d) reached",
+	    @"Force disconnect of '%@' because max connections (%u) reached",
 	    other, maxConnections]; 
 	}
       [other disconnect];
@@ -1180,6 +1006,7 @@ static unsigned int	maxConnections = 8;
                     }
                 }
 
+	      _lastStart = GSTickerTimeNow();
 	      [self backendConnect];
               /* On establishng a new connection, we must restore any
                * listen instructions in the backend.
@@ -1233,15 +1060,17 @@ static unsigned int	maxConnections = 8;
 {
   NSNotificationCenter	*nc;
 
-  if (_name != nil)
+  [clientsLock lock];
+  NSHashRemove(clientsHash, (void*)self);
+  if (_name != nil
+    && (SQLClient*)NSMapGet(clientsMap, (void*)_name) == self)
     {
-      [clientsMapLock lock];
       NSMapRemove(clientsMap, (void*)_name);
-      [clientsMapLock unlock];
     }
+  [clientsLock unlock];
   nc = [NSNotificationCenter defaultCenter];
   [nc removeObserver: self];
-  [self disconnect];
+  if (YES == connected) [self disconnect];
   [lock release]; lock = nil;
   [_client release]; _client = nil;
   [_database release]; _database = nil;
@@ -1381,6 +1210,14 @@ static unsigned int	maxConnections = 8;
 
 - (id) initWithConfiguration: (NSDictionary*)config
 			name: (NSString*)reference
+{
+  return [self initWithConfiguration: config name: reference pool: nil];
+}
+
+
+- (id) initWithConfiguration: (NSDictionary*)config
+			name: (NSString*)reference
+                        pool: (SQLClientPool*)pool
 {
   NSNotification	*n;
   NSDictionary		*conf = config;
@@ -1401,8 +1238,16 @@ static unsigned int	maxConnections = 8;
 	}
     }
 
-  [clientsMapLock lock];
-  existing = (SQLClient*)NSMapGet(clientsMap, reference);
+  [clientsLock lock];
+  _pool = pool;
+  if (nil != _pool)
+    {
+      existing = nil;   // Pool, object ... can't already exist
+    }
+  else
+    {
+      existing = (SQLClient*)NSMapGet(clientsMap, reference);
+    }
   if (nil == existing)
     {
       lock = [NSRecursiveLock new];	// Ensure thread-safety.
@@ -1426,6 +1271,7 @@ static unsigned int	maxConnections = 8;
 	object: conf
 	userInfo: nil];
 
+      NSHashInsert(clientsHash, (void*)self);
       [self _configure: n];	// Actually set up the configuration.
     }
   else
@@ -1433,7 +1279,7 @@ static unsigned int	maxConnections = 8;
       [self release];
       self = [existing retain];
     }
-  [clientsMapLock unlock];
+  [clientsLock unlock];
 
   return self;
 }
@@ -1478,6 +1324,51 @@ static unsigned int	maxConnections = 8;
   return nil;
 }
 
+- (SQLClient*) longestIdle: (SQLClient*)other
+{
+  NSTimeInterval        t0;
+  NSTimeInterval        t1;
+
+  NSAssert([other isKindOfClass: SQLClientClass], NSInvalidArgumentException);
+
+  t0 = _lastOperation;
+  if (t0 < _lastStart)
+    {
+      t0 = _lastStart;
+    }
+  if (NO == connected || 0 != _connectFails)
+    {
+      t0 = 0.0;
+    }
+
+  if (YES == [other isProxy])
+    {
+      t1 = 0.0;
+    }
+  else
+    {
+      t1 = other->_lastOperation;
+      if (t1 < other->_lastStart)
+        {
+          t1 = other->_lastStart;
+        }
+      if (NO == connected || 0 != other->_connectFails)
+        {
+          t1 = 0.0;
+        }
+    }
+
+  if (t0 <= 0.0 && t1 <= 0.0)
+    {
+      return nil;
+    }
+  if (t1 <= t0)
+    {
+      return other;
+    }
+  return self;
+}
+
 - (NSString*) name
 {
   return _name;
@@ -1733,12 +1624,19 @@ static unsigned int	maxConnections = 8;
    * it from the table so that no other thread will find it
    * and try to use it while it is being deallocated.
    */
-  [clientsMapLock lock];
+  [clientsLock lock];
   if (NSDecrementExtraRefCountWasZero(self))
     {
-      [self dealloc];
+      if (nil != _pool)
+        {
+          [_pool swallowClient: self];
+        }
+      else
+        {
+          [self dealloc];
+        }
     }
-  [clientsMapLock unlock];
+  [clientsLock unlock];
 }
 
 - (void) rollback
@@ -1778,7 +1676,7 @@ static unsigned int	maxConnections = 8;
     {
       if ([s isEqual: _database] == NO)
         {
-          if (connected == YES)
+          if (YES == connected)
             {
               [self disconnect];
             }
@@ -1803,24 +1701,28 @@ static unsigned int	maxConnections = 8;
     {
       if ([s isEqual: _name] == NO)
         {
-          [clientsMapLock lock];
-          if (NSMapGet(clientsMap, s) != 0)
+          [clientsLock lock];
+          if (nil == _pool)
             {
-              [clientsMapLock unlock];
-              [lock unlock];
-              if ([self debugging] > 0)
+              if (NSMapGet(clientsMap, s) != 0)
                 {
-                  [self debug: @"Error attempt to re-use client name %@", s];
+                  [clientsLock unlock];
+                  [lock unlock];
+                  if ([self debugging] > 0)
+                    {
+                      [self
+                        debug: @"Error attempt to re-use client name %@", s];
+                    }
+                  NS_VOIDRETURN;
                 }
-              NS_VOIDRETURN;
             }
-          if (connected == YES)
+          if (YES == connected)
             {
               [self disconnect];
             }
-          if (_name != nil)
+          if (_name != nil
+            && (SQLClient*)NSMapGet(clientsMap, (void*)_name) == self)
             {
-              [[self retain] autorelease];
               NSMapRemove(clientsMap, (void*)_name);
             }
           s = [s copy];
@@ -1828,8 +1730,11 @@ static unsigned int	maxConnections = 8;
           _name = s;
           [_client release];
           _client = [[[NSProcessInfo processInfo] globallyUniqueString] retain];
-          NSMapInsert(clientsMap, (void*)_name, (void*)self);
-          [clientsMapLock unlock];
+          if (nil == _pool && _name != nil)
+            {
+              NSMapInsert(clientsMap, (void*)_name, (void*)self);
+            }
+          [clientsLock unlock];
         }
     }
   NS_HANDLER
@@ -1848,7 +1753,7 @@ static unsigned int	maxConnections = 8;
     {
       if ([s isEqual: _password] == NO)
         {
-          if (connected == YES)
+          if (YES == connected)
             {
               [self disconnect];
             }
@@ -1876,9 +1781,9 @@ static unsigned int	maxConnections = 8;
   [lock lock];
   NS_DURING
     {
-      if ([s isEqual: _client] == NO)
+      if ([s isEqual: _user] == NO)
         {
-          if (connected == YES)
+          if (YES == connected)
             {
               [self disconnect];
             }
@@ -1904,7 +1809,6 @@ static unsigned int	maxConnections = 8;
   [lock lock];
   NS_DURING
     {
-      NSTimeInterval	start = 0.0;
       NSString	        *statement;
       BOOL              isCommit = NO;
       BOOL              isRollback = NO;
@@ -1920,10 +1824,7 @@ static unsigned int	maxConnections = 8;
           isRollback = YES;
         }
 
-      if (_duration >= 0)
-	{
-	  start = GSTickerTimeNow();
-	}
+      _lastStart = GSTickerTimeNow();
       result = [self backendExecute: info];
       _lastOperation = GSTickerTimeNow();
       [_statements addObject: statement];
@@ -1931,7 +1832,7 @@ static unsigned int	maxConnections = 8;
 	{
 	  NSTimeInterval	d;
 
-	  d = _lastOperation - start;
+	  d = _lastOperation - _lastStart;
 	  if (d >= _duration)
 	    {
 	      if (isCommit || isRollback)
@@ -2012,19 +1913,14 @@ static unsigned int	maxConnections = 8;
   [lock lock];
   NS_DURING
     {
-      NSTimeInterval	start = 0.0;
-
-      if (_duration >= 0)
-	{
-	  start = GSTickerTimeNow();
-	}
+      _lastStart = GSTickerTimeNow();
       result = [self backendQuery: stmt recordType: rtype listType: ltype];
       _lastOperation = GSTickerTimeNow();
       if (_duration >= 0)
 	{
 	  NSTimeInterval	d;
 
-	  d = _lastOperation - start;
+	  d = _lastOperation - _lastStart;
 	  if (d >= _duration)
 	    {
 	      debug = [NSString stringWithFormat:
@@ -2187,11 +2083,6 @@ static unsigned int	maxConnections = 8;
 
 @implementation	SQLClient (Private)
 
-/**
- * Internal method to handle configuration using the notification object.
- * This object may be either a configuration front end or a user defaults
- * object ... so we have to be careful that we work with both.
- */
 - (void) _configure: (NSNotification*)n
 {
   NSDictionary	*o;
@@ -2242,7 +2133,8 @@ static unsigned int	maxConnections = 8;
             {
               path = [paths objectAtIndex: count];
               path = [path stringByAppendingPathComponent: @"Bundles"];
-              path = [path stringByAppendingPathComponent: @"SQLClient"];
+              path = [path stringByAppendingPathComponent:
+			     @"SQLClient"SOVERSION""];
               path = [path stringByAppendingPathComponent: s];
               path = [path stringByAppendingPathExtension: @"bundle"];
               bundle = [NSBundle bundleWithPath: path];
@@ -2293,7 +2185,10 @@ static unsigned int	maxConnections = 8;
         }
       if (c != [self class])
         {
-          [self disconnect];
+          if (YES == connected)
+            {
+              [self disconnect];
+            }
 #ifdef	GNUSTEP
           GSDebugAllocationRemove(object_getClass(self), self);
 #endif
@@ -2345,8 +2240,11 @@ static unsigned int	maxConnections = 8;
   [lock unlock];
 }
 
-/** Internal method to populate the cache with the result of a query.
- */
+- (NSRecursiveLock*) _lock
+{
+  return lock;
+}
+
 - (void) _populateCache: (CacheQuery*)a
 {
   GSCache	*cache;
@@ -2374,14 +2272,7 @@ static unsigned int	maxConnections = 8;
 	  lifetime: a->lifetime];
 }
 
-/**
- * Internal method to build an sql string by quoting any non-string objects
- * and concatenating the resulting strings in a nil terminated list.<br />
- * Returns an array containing the statement as the first object and
- * any NSData objects following.  The NSData objects appear in the
- * statement strings as the marker sequence - <code>'?'''?'</code>
- */
-- (NSArray*) _prepare: (NSString*)stmt args: (va_list)args
+- (NSMutableArray*) _prepare: (NSString*)stmt args: (va_list)args
 {
   NSMutableArray	*ma = [NSMutableArray arrayWithCapacity: 2];
   NSString		*tmp = va_arg(args, NSString*);
@@ -2427,15 +2318,7 @@ static unsigned int	maxConnections = 8;
   mainThread = [NSThread currentThread];
 }
 
-/**
- * Internal method to substitute values from the dictionary into
- * a string containing markup identifying where the values should
- * appear by name.  Non-string objects in the dictionary are quoted.<br />
- * Returns an array containing the statement as the first object and
- * any NSData objects following.  The NSData objects appear in the
- * statement strings as the marker sequence - <code>'?'''?'</code>
- */
-- (NSArray*) _substitute: (NSString*)str with: (NSDictionary*)vals
+- (NSMutableArray*) _substitute: (NSString*)str with: (NSDictionary*)vals
 {
   unsigned int		l = [str length];
   NSRange		r;
@@ -2609,10 +2492,6 @@ static unsigned int	maxConnections = 8;
   return ma;
 }
 
-/*
- * Called at one second intervals to ensure that our current timestamp
- * is reasonably accurate.
- */
 + (void) _tick: (NSTimer*)t
 {
   (void) GSTickerTimeNow();
@@ -2876,7 +2755,6 @@ static unsigned int	maxConnections = 8;
 {
   NSMutableArray	*result;
   NSMutableDictionary	*md;
-  NSTimeInterval	start;
   GSCache		*c;
   id			toCache;
 
@@ -2886,7 +2764,7 @@ static unsigned int	maxConnections = 8;
   md = [[NSThread currentThread] threadDictionary];
   [md setObject: rtype forKey: @"SQLClientRecordType"];
   [md setObject: ltype forKey: @"SQLClientListType"];
-  start = GSTickerTimeNow();
+  _lastStart = GSTickerTimeNow();
   c = [self cache];
   toCache = nil;
 
@@ -2931,7 +2809,7 @@ static unsigned int	maxConnections = 8;
 	{
 	  NSTimeInterval	d;
 
-	  d = _lastOperation - start;
+	  d = _lastOperation - _lastStart;
 	  if (d >= _duration)
 	    {
 	      [self debug: @"Duration %g for query %@", d, stmt];
@@ -3093,20 +2971,184 @@ static unsigned int	maxConnections = 8;
     }
 }
 
+/* Try to merge the prepared statement p with an earlier statement in the
+ * transaction.  We search up to 5 earlier statements and we merge if we can.
+ */
+- (void) _merge: (NSMutableArray*)p
+{
+  if (_count > 0 && _merge > 0)
+    {
+      static NSCharacterSet     *w = nil;
+      NSString  *s;
+      NSRange   r;
+
+      s = [p objectAtIndex: 0];         // Get SQL part of array
+
+      if (nil == w)
+        {
+          w = [[NSCharacterSet whitespaceAndNewlineCharacterSet] retain];
+        }
+
+      r = [s rangeOfString: @"INSERT" options: NSCaseInsensitiveSearch];
+      if (r.length > 0 && 0 == r.location)
+        {
+          r = [s rangeOfString: @"VALUES" options: NSCaseInsensitiveSearch];
+          if (r.length > 0)
+            {
+              NSUInteger        l = [s length];
+              NSUInteger        pos = NSMaxRange(r);
+
+              while (pos < l
+                && [w characterIsMember: [s characterAtIndex: pos]])
+                {
+                  pos++;
+                }
+              if (pos < l && [s characterAtIndex: pos] == '(')
+                {
+                  NSString              *t = [s substringToIndex: pos];
+                  NSUInteger            index = _count;
+                  NSUInteger            attempts = 0;
+
+                  s = [s substringFromIndex: pos];
+                  while (index-- > 0 && attempts++ < _merge)
+                    {
+                      NSMutableArray    *o;
+                      NSString          *os;
+
+                      o = [_info objectAtIndex: index];
+                      os = [o objectAtIndex: 0];
+                      if ([os hasPrefix: t])
+                        {
+                          NSMutableString       *m;
+
+                          if ([os isKindOfClass: [NSMutableString class]])
+                            {
+                              m = (NSMutableString*)os;
+                            }
+                          else
+                            {
+                              m = [NSMutableString
+                                stringWithCapacity: [os length] * 100];
+                              [m appendString: os];
+                            }
+                          [m appendString: @","];
+                          [m appendString: s];
+                          [o replaceObjectAtIndex: 0 withObject: m];
+                          for (index = 1; index < [p count]; index++)
+                            {
+                              [o addObject: [p objectAtIndex: index]];
+                            }
+                          return;
+                        }
+                    }
+                }
+            }
+        }
+
+      r = [s rangeOfString: @"UPDATE" options: NSCaseInsensitiveSearch];
+      if (0 == r.length)
+        {
+          r = [s rangeOfString: @"DELETE" options: NSCaseInsensitiveSearch];
+        }
+      if (r.length > 0 && 0 == r.location)
+        {
+          r = [s rangeOfString: @"WHERE" options: NSCaseInsensitiveSearch];
+          if (r.length > 0)
+            {
+              NSUInteger        l = [s length];
+              NSUInteger        pos = NSMaxRange(r);
+
+              while (pos < l
+                && [w characterIsMember: [s characterAtIndex: pos]])
+                {
+                  pos++;
+                }
+              if (pos < l && [s characterAtIndex: pos] == '(')
+                {
+                  NSString              *t = [s substringToIndex: pos];
+                  NSUInteger            index = _count;
+                  NSUInteger            attempts = 0;
+
+                  /* Get the condition after the WHERE and if it's not
+                   * in brackets, add them so the merge can work.
+                   */
+                  s = [s substringFromIndex: pos];
+                  if ([s characterAtIndex: 0] != '(')
+                    {
+                      s = [NSString stringWithFormat: @"(%@)", s];
+                    }
+                    
+                  while (index-- > 0 && attempts++ < _merge)
+                    {
+                      NSMutableArray    *o;
+                      NSString          *os;
+
+                      o = [_info objectAtIndex: index];
+                      os = [o objectAtIndex: 0];
+                      if ([os hasPrefix: t])
+                        {
+                          NSMutableString       *m;
+
+                          l = [os length];
+                          if ([os characterAtIndex: l - 1] == ')')
+                            {
+                              if ([os isKindOfClass: [NSMutableString class]])
+                                {
+                                  m = (NSMutableString*)os;
+                                }
+                              else
+                                {
+                                  m = [NSMutableString
+                                    stringWithCapacity: l * 100];
+                                  [m appendString: os];
+                                }
+                            }
+                          else
+                            {
+                              /* The condition of the WHERE clause was not
+                               * bracketed, so we extract it and build a
+                               * new statement in which it is bracketed.
+                               */
+                              os = [os substringFromIndex: pos];
+                              m = [NSMutableString
+                                stringWithCapacity: l * 100];
+                              [m appendFormat: @"%@(%@)", t, os];
+                            }
+                          [m appendString: @" OR "];
+                          [m appendString: s];
+                          [o replaceObjectAtIndex: 0 withObject: m];
+                          for (index = 1; index < [p count]; index++)
+                            {
+                              [o addObject: [p objectAtIndex: index]];
+                            }
+                          return;
+                        }
+                    }
+                }
+            }
+        }
+    }
+  [_info addObject: p];
+  _count++;
+}
+
 - (void) add: (NSString*)stmt,...
 {
-  va_list       ap;
+  va_list               ap;
+  NSMutableArray        *p;
 
   va_start (ap, stmt);
-  [_info addObject: [_db _prepare: stmt args: ap]];
-  _count++;
+  p = [_db _prepare: stmt args: ap];
   va_end (ap);
+  [self _merge: p];
 }
 
 - (void) add: (NSString*)stmt with: (NSDictionary*)values
 {
-  [_info addObject: [_db _substitute: stmt with: values]];
-  _count++;
+  NSMutableArray        *p;
+
+  p = [_db _substitute: stmt with: values];
+  [self _merge: p];
 }
 
 - (void) append: (SQLTransaction*)other
@@ -3119,10 +3161,25 @@ static unsigned int	maxConnections = 8;
 		      format: @"[%@-%@] database client missmatch",
 	    NSStringFromClass([self class]), NSStringFromSelector(_cmd)];
 	}
-      other = [other copy];
-      [_info addObject: other];
-      _count += other->_count;
-      [other release];
+      if (_merge > 0)
+        {
+          unsigned      index;
+
+          /* Merging of statements is turned on ... try to merge statements
+           * from other transaction rather than simply appending a copy of it.
+           */
+          for (index = 0; index < other->_count; index++)
+            {
+              [self _merge: [other->_info objectAtIndex: index]];
+            }
+        }
+      else
+        {
+          other = [other copy];
+          [_info addObject: other];
+          _count += other->_count;
+          [other release];
+        }
     }
 }
 
@@ -3165,7 +3222,11 @@ static unsigned int	maxConnections = 8;
   if (_count > 0)
     {
       NSMutableArray    *info = nil;
+      NSRecursiveLock   *dbLock = [_db _lock];
+      BOOL              wrap;
 
+      [dbLock lock];
+      wrap = [_db isInTransaction] ? NO : YES;
       NS_DURING
 	{
           NSMutableString   *sql;
@@ -3180,25 +3241,42 @@ static unsigned int	maxConnections = 8;
           sql = [[NSMutableString alloc] initWithCapacity: sqlSize + 13];
           [info addObject: sql];
           [sql release];
-          if ([_db isInTransaction] == NO)
+          if (YES == wrap)
             {
               [sql appendString: @"begin;"];
             }
 
           [self _addSQL: sql andArgs: info];
 
-          if ([_db isInTransaction] == NO)
+          if (YES == wrap)
             {
               [sql appendString: @"commit;"];
             }
 
           [_db simpleExecute: info];
           [info release]; info = nil;
+          [dbLock unlock];
 	}
       NS_HANDLER
 	{
+          NSException   *e = localException;
+
           [info release];
-	  [localException raise];
+          if (YES == wrap)
+            {
+              NS_DURING
+                {
+                  [_db simpleExecute: rollbackStatement];
+                }
+              NS_HANDLER
+                {
+                  [_db disconnect];
+                  NSLog(@"Disconnected due to failed rollback after %@", e);
+                }
+              NS_ENDHANDLER
+            }
+          [dbLock unlock];
+          [e raise];
 	}
       NS_ENDHANDLER
     }
@@ -3216,6 +3294,9 @@ static unsigned int	maxConnections = 8;
 
   if (_count > 0)
     {
+      NSRecursiveLock   *dbLock = [_db _lock];
+
+      [dbLock lock];
       NS_DURING
         {
           [self execute];
@@ -3318,6 +3399,7 @@ static unsigned int	maxConnections = 8;
             }
         }
       NS_ENDHANDLER
+      [dbLock unlock];
     }
   return executed;
 }
@@ -3376,6 +3458,14 @@ static unsigned int	maxConnections = 8;
   _count = 0;
 }
 
+- (uint8_t) setMerge: (uint8_t)history
+{
+  uint8_t       old = _merge;
+
+  _merge = history;
+  return old;
+}
+
 - (unsigned) totalCount
 {
   return _count;
@@ -3695,3 +3785,22 @@ validName(NSString *name)
 }
 @end
 
+@implementation	SQLClientPool (Adjust)
+
++ (void) _adjustPoolConnections: (int)n
+{
+  unsigned      err = 0;
+
+  [clientsLock lock];
+  poolConnections += n;
+  if (poolConnections < 0)
+    {
+      err -= poolConnections;
+      poolConnections = 0;
+    }
+  [clientsLock unlock];
+  NSAssert(0 == err, NSInvalidArgumentException);
+}
+
+@end
+
